package pl.edu.uw.heroes.simulation;

import lombok.AllArgsConstructor;
import pl.edu.uw.heroes.actions.Action;
import pl.edu.uw.heroes.actions.Defend;
import pl.edu.uw.heroes.actions.Move;
import pl.edu.uw.heroes.actions.Wait;
import pl.edu.uw.heroes.board.BFSCalculator;
import pl.edu.uw.heroes.units.Unit;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Queue;
import java.util.stream.Stream;

@AllArgsConstructor
public class GameSimulation {

    private final GameState state;

    private final BFSCalculator bfsCalculator = new BFSCalculator();

    public void executeOneAction() {
        Queue<Unit> units = state.getUnits();
        if (units.isEmpty())
            prepareNextRound();

        Unit unit = units.poll();

        Collection<Action> actions = actionsForUnit(unit);
        Action action = unit.getOwner().chooseAction(actions);
        System.out.println(unit.getOwner() + " chose " + action);
        action.execute(state);
        state.getExecutedActions().add(action);
    }

    private void prepareNextRound() {
        Queue<Unit> units = state.getUnits();
        Stream.concat(
                        state.getPlayerLeft().getUnits().stream(),
                        state.getPlayerRight().getUnits().stream()
                )
                .peek(Unit::resetRound)
                .sorted((u1, u2) -> Integer.compare(u2.getSpeed(), u1.getSpeed()))
                .forEach(units::add);
        if (units.isEmpty())
            throw new IllegalStateException("No units on board!");
    }

    private Collection<Action> actionsForUnit(Unit unit) {
        Collection<Action> actions = new ArrayList<>();
        actions.add(new Defend(unit));
        if (unit.canWait())
            actions.add(new Wait(unit));
        bfsCalculator.calculatePossibleMoves(unit.getField()).forEach(field -> actions.add(new Move(unit, field)));
        return actions;
    }

    public void printState() {
        System.out.println("Player left (" + state.getPlayerLeft() + "): ");
        System.out.println(state.getPlayerLeft().getUnits());
        System.out.println("Player right (" + state.getPlayerRight() + "): ");
        System.out.println(state.getPlayerRight().getUnits());
    }
}
